from pathlib import Path
import duckdb
import pyarrow as pa
from texture.models import (
    DuckQueryData,
    DuckQueryResult,
    ExecResponse,
    JsonResponse,
    ErrorResponse,
)
from fastapi.responses import Response
from texture.database.example_data import (
    EXAMPLE_DATASET_INFO,
    EXAMPLE_DATA_PATHS,
)
import pandas as pd
import numpy as np
import lancedb
import torch
import sentence_transformers

# NOTE: this path is affected by where the server is run from, assuming it is run in the root for now
CACHE_DIR = ".texture_cache/"
CACHE_PATH = Path(CACHE_DIR)
CACHE_PATH.mkdir(parents=True, exist_ok=True)


def get_embedding(col: np.ndarray, model_name):
    model = sentence_transformers.SentenceTransformer(model_name)
    e = model.encode(col)
    return e


def load_datasets(duckdbconn, dsGroup):
    for dsName, dsPath in EXAMPLE_DATA_PATHS[dsGroup]["datasets"].items():
        duckdbconn.load_dataset(dsName, dsPath)


def load_embeddings(vectordbconn, dsGroup):
    for dsName, embeddingPath in EXAMPLE_DATA_PATHS[dsGroup]["embeddings"].items():
        embeddings = torch.load(CACHE_PATH / embeddingPath)
        embeddings = (
            embeddings.cpu().numpy()
            if isinstance(embeddings, torch.Tensor)
            else embeddings
        )
        df = pd.read_parquet(
            CACHE_PATH / EXAMPLE_DATA_PATHS[dsGroup]["datasets"][dsName]
        )

        embed_func = lambda x: get_embedding(
            x, "all-mpnet-base-v2"
        )  # TODO change this model to be the same one used for making embedding

        vectordbconn.add_table(dsName, df, embeddings, "id", embed_func)


def populate_example_datasets(duckdbconn, vectordbconn, metadataCache):
    print("Loading example datasets...")
    for dsGroupName in EXAMPLE_DATASET_INFO:
        print(f"Loading data for {dsGroupName}...")
        load_datasets(duckdbconn, dsGroupName)
        if "embeddings" in EXAMPLE_DATA_PATHS[dsGroupName]:
            load_embeddings(vectordbconn, dsGroupName)
        metadataCache[dsGroupName] = EXAMPLE_DATASET_INFO[dsGroupName]


def populate_dataset(duckdb_conn, vectordb_conn, datasetMetadataCache, ds_init_info):
    print("Loading dataset: ", ds_init_info.datasetInfo.name)
    datasetMetadataCache[ds_init_info.datasetInfo.name] = ds_init_info.datasetInfo
    for table_name, table_df in ds_init_info.load_tables.items():
        duckdb_conn.load_dataframe(table_name, table_df)

    if ds_init_info.load_embeddings:
        for table_name, embedding_matrix in ds_init_info.load_embeddings.items():

            table_df = ds_init_info.load_tables[table_name]

            embed_func = lambda x: get_embedding(
                x, "all-mpnet-base-v2"
            )  # TODO change this model to be the same one used for making embedding

            vectordb_conn.add_table(
                table_name,
                table_df,
                embedding_matrix,
                ds_init_info.datasetInfo.primary_key.name,
                embed_func,
            )


def init_db():
    duckdbconn = DatabaseConnection()
    vectordbconn = VectorDBConnection()
    metadataCache = {}

    # print("Texture: database initialized.")
    return duckdbconn, vectordbconn, metadataCache


class DatabaseConnection:
    def __init__(self, database_name="defaultDatabase.db"):
        # Path(CACHE_PATH).mkdir(parents=True, exist_ok=True)
        # p = Path(CACHE_PATH) / database_name

        # NOTE: can save this to file, but potentially causes issues with old tables
        # so right now making new database on start up each time
        self.connection = duckdb.connect()

    def query(self, query_string: str) -> list:
        """
        Executes SQL query on database
        See other return types: https://duckdb.org/docs/archive/0.9.2/api/python/overview

        Returns:
            - Array of python objects representing rows
        """
        return self.connection.sql(query_string).fetchall()

    def execute(self, query_string: str) -> None:
        """
        Executes SQL that does not expect return.

        Returns:
            - None
        """
        self.connection.sql(query_string)

    def load_dataset(self, dataset_name: str, dataset_path: str):
        """
        Loads a dataset into the database

        Args:
            dataset_name: name of the dataset
            dataset_path: path to parquet file
        """
        self.execute(
            f"CREATE TABLE IF NOT EXISTS '{dataset_name}' AS (SELECT * FROM read_parquet('{CACHE_PATH / dataset_path}'));"
        )

    def load_dataframe(self, table_name, df: pd.DataFrame):
        """
        Loads a DataFrame into the database

        Args:
            dataset_name: name of the dataset
            df: DataFrame to load
        """
        q = f"CREATE TABLE '{table_name}' AS SELECT * FROM df"
        self.execute(q)

    def add_column(self, tableName, columnName, data):
        """
        Add a new column to the table with the given data. Data must match the length of the table.

        Args:
            tableName: name of the dataset
            columnName: NEW column name to add
            data: data to add. Can be list-like or pd.Series
        """

        # get current data
        current_df = self.connection.execute(f"SELECT * FROM {tableName}").fetchdf()
        current_df[columnName] = data

        # modify existing table
        viewName = f"{tableName}_TEMP_NEW"

        self.connection.register(viewName, current_df)
        self.connection.execute(
            f"CREATE OR REPLACE TABLE {tableName} as SELECT * FROM {viewName}"
        )
        self.connection.execute(f"DROP VIEW IF EXISTS {viewName}")

    def write_table_to_file(self, table_name, file_path):
        self.connection.execute(
            f"""COPY (SELECT * FROM "{table_name}") TO '{CACHE_PATH / file_path}' (FORMAT 'parquet');"""
        )

    def _handle_json_message(self, data: DuckQueryData) -> DuckQueryResult:
        """
        From: https://github.com/uwdata/mosaic/blob/main/packages/widget/mosaic_widget/__init__.py
        """
        uuid = data.uuid
        sql = data.sql
        response_message = None

        try:
            if data.type == "exec":
                self.connection.execute(sql)
                response_message = ExecResponse(type="exec", uuid=uuid)
            elif data.type == "json":
                query_result = self.connection.query(sql).df()
                json = query_result.to_dict(orient="records")

                response_message = JsonResponse(type="json", uuid=uuid, result=json)

            else:
                response_message = ErrorResponse(
                    error=f"Unsupported response type: {data.type}", uuid=uuid
                )

        except Exception as e:
            print("[Handle JSON message ERROR]: ", e, "\nExecuting SQL: ", sql)
            response_message = ErrorResponse(error=str(e), uuid=uuid)

        return response_message

    def _handle_arrow_message(self, data: DuckQueryData):
        """
        Arrow handler. Different function since returns Response; might be able to coalese in future
        """
        uuid = data.uuid
        sql = data.sql
        response_message = None

        try:
            if data.type == "arrow":
                query_result = self.connection.query(sql).arrow()

                sink = pa.BufferOutputStream()

                with pa.ipc.new_stream(sink, query_result.schema) as writer:
                    writer.write(query_result)

                pybytes = sink.getvalue().to_pybytes()
                response_message = Response(
                    content=pybytes,
                    media_type="application/vnd.apache.arrow.stream",
                )

            else:
                response_message = ErrorResponse(
                    error=f"Unsupported response type: {data.type}", uuid=uuid
                )
        except Exception as e:
            print("[Handle Arrow message ERROR]: ", e, "\nExecuting SQL: ", sql)
            response_message = ErrorResponse(error=str(e), uuid=uuid)

        return response_message


class VectorDBConnection:
    def __init__(self, database_dir="lancedb"):
        """
        embed_func: function that takes a string and returns a numpy array embedding.
        This MUST be same model as original embeddings or else comparison will not work
        """

        self.connection = lancedb.connect(CACHE_PATH / database_dir)
        self.id_cols = {}
        self.embed_funcs = {}

    def _check(self, table_name, check_conn=False, check_id=False, check_embed=False):
        if check_conn and table_name not in self.connection:
            raise ValueError(f"Table {table_name} not found in LanceDB.")

        if check_id and table_name not in self.id_cols:
            raise ValueError(f"Table {table_name} does not have an id column saved.")

        if check_embed and table_name not in self.embed_funcs:
            raise ValueError(f"Table {table_name} does not have an embedding function.")

    def add_table(
        self,
        table_name: str,
        data: pd.DataFrame,
        embeddings: np.ndarray,
        id_col_name: str,
        embed_func,
    ):
        """
        Add a table to the database.

        Args:
            table_name (str): The name of the table.
            data (pd.DataFrame): The dataframe containing the metadata and a column called "vector" with numpy arrays per row representing the vector representation.
            id_col_name (str): The name of the column in the dataframe that contains the unique identifier for each row.
            embed_func: function that takes a string and returns a numpy array embedding for this table

        Returns:
            None
        """
        data["vector"] = list(embeddings)
        self.table = self.connection.create_table(
            table_name, data=data, mode="overwrite"
        )
        self.id_cols[table_name] = id_col_name
        self.embed_funcs[table_name] = embed_func

    def search(self, table_name: str, vector: np.array, limit: int = 20):
        """
        Find ids of KNN docs to vector
        """
        self._check(table_name, check_conn=True, check_id=True)

        id_col = self.id_cols[table_name]
        result = (
            self.connection[table_name].search(vector).limit(limit).select([id_col])
        )

        return result.to_pandas()[[id_col, "_distance"]]

    def get_embedding_from_string(self, table_name, text: str) -> np.array:
        self._check(table_name, check_embed=True)

        return self.embed_funcs[table_name](text)

    def get_embedding_from_id(self, table_name, id: str) -> np.array:
        """
        Get the embedding from the table by id
        """
        self._check(table_name, check_conn=True, check_id=True)

        df = (
            self.connection[table_name]
            .search()
            .where(f"{self.id_cols[table_name]} = {id}")
            .to_pandas()["vector"]
        )

        return df.iloc[0]



from functools import wraps
import time
from texture.models import DataType
import pandas as pd


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        # print(f"Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds")
        print(f"{func.__name__} Took {total_time:.4f} seconds")
        return result

    return timeit_wrapper


def get_type_from_response(inputType) -> DataType:
    """input type is number, string, or bool from llm response format"""
    if inputType == "number":
        return "number"

    return "categorical"


def flatten(col: pd.Series, idColName="id"):
    """flatten a series with arrays into a dataframe with one entry per row"""
    col = col.explode()
    col = col.reset_index()
    col = col.rename(columns={"index": idColName})

    return col


def is_notebook() -> bool:
    try:
        from IPython.core.getipython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except (NameError, ImportError):
        return False  # Probably standard Python interpreter



import pandas as pd


def execute_code_and_apply_function(code_str: str, series: pd.Series) -> pd.Series:
    """
    Execute the code string and apply the function to the series.

    Args:
        code_str (str): The code string. Should define a function called transform.
        series (pd.Series): The series to apply the function to.
    """

    # use the same dict for local and global context so that helper functions can be accessed
    # FUTURE: probably want to use this for imports like context = {"pd": pd}
    context = {}

    exec(code_str, context, context)
    t = context["transform"]
    result = t(series)

    return result



import pandas as pd
from nltk.tokenize import TreebankWordTokenizer as twt
from concurrent.futures import ThreadPoolExecutor

twt_tokenizer = twt()


def get_words_w_span(text):
    if not text:
        return pd.DataFrame(columns=["span_start", "span_end", "word"])

    token_idx = twt_tokenizer.span_tokenize(text)
    _df = pd.DataFrame(token_idx, columns=["span_start", "span_end"])
    _df["word"] = _df.apply(lambda x: text[x.span_start : x.span_end], axis=1)

    return _df


def get_words_w_span_batch(arr: list[str], num_threads=8) -> list[list[str]]:
    with ThreadPoolExecutor(num_threads) as executor:
        return list(executor.map(get_words_w_span, arr))


def get_df_words_w_span(arr: list, id_col: list):
    """
    Calculate the words and their location.

    Args:
        arr (list): List of strings to tokenize.
        id_col (list): List of document ids (must be same size as arr)
    """
    result_list = get_words_w_span_batch(arr)

    for idx, _mydf in zip(id_col, result_list):
        _mydf["id"] = idx

    all_d = pd.concat(result_list, ignore_index=True)
    all_d["word"] = all_d["word"].str.lower()
    return all_d



import torch
import pandas as pd
from typing import List
from texture.models import ColumnInputInfo


def get_data_types(_df: pd.DataFrame) -> List[ColumnInputInfo]:
    """
    Get the data types of each column in the DataFrame.
    Maps each column to a semantic type: 'number', 'date', 'text', or 'categorical'
    """
    data_info = []
    for col_name, dtype in _df.dtypes.items():
        inferred_type = None

        if pd.api.types.is_numeric_dtype(dtype):
            inferred_type = "number"
        elif pd.api.types.is_datetime64_any_dtype(dtype):
            inferred_type = "date"
        else:
            max_col_len = _df[col_name].str.len().max()
            cardinality = _df[col_name].nunique()

            # heuristic to detect text or categorical
            if (max_col_len > 200) or (cardinality == len(_df)):
                inferred_type = "text"
            else:
                inferred_type = "categorical"

        c = ColumnInputInfo(name=col_name, type=inferred_type)
        data_info.append(c)
    return data_info


def save_embeddings(embedding, path):
    torch.save(embedding, path)
    print("Saved embedding to ", path)


def load_embeddings(path):
    e = torch.load(path)
    print("Loaded embedding from ", path)
    return e



from texture.preprocess.utils import get_data_types, save_embeddings, load_embeddings
from texture.preprocess.embeddings import (
    get_embeddings_and_projection,
    get_projection,
    calculate_embeddings,
)
from texture.preprocess.tokenize import get_df_words_w_span
from texture.preprocess.pipeline import validate_and_run_preprocess



import openai
from openai import AsyncOpenAI
import json
import asyncio
from tqdm.asyncio import tqdm_asyncio
from typing import List
import tiktoken

from texture.utils import timeit
from texture.models import TaskFormat

# Note - json mode only works with these two model names
# MODEL_NAME = "gpt-4-turbo-preview"
# MAX_TPM = 450000
MODEL_NAME = "gpt-3.5-turbo"
MAX_TPM = 80000
# MODEL_NAME = "gpt-3.5-turbo-instruct"  # higher token limit but have to coerce to valid json
model_encoding = tiktoken.encoding_for_model(MODEL_NAME)

MAX_BATCH_SIZE = 150


################ PROMPT 1: For getting the json representation of the response format
PROMPT_GET_JSON_FORMAT = """# Instruction
You are a data processing assistant. You will be helping a user transform a text column by extracting the requested information and turning it into new columns. To do this, we need to know the JSON specification of the requested columns. Given a user instruction, return the following information about the columns that will be generated to help answer that question.

# Response format
The response should be an object with three keys:
1. "name": the column name
1. "type": the response type of either "number", "string", or "bool"
2. "num_replies": if the response will be a "single" item or "multiple" items

# Examples
-----
- Instruction: 
Extract all email addresses from the given text string.
- JSON format: 
{ "name": "email_addresses", "type": "string", "num_replies": "multiple" }
-----
- Instruction: 
Summarize this document
- JSON format: 
{ "name": "summary", "type": "string", "num_replies": "single" }
-----
- Instruction: 
Extract the salary from this job posting
- JSON format: 
{ "name": "salaries", "type": "number", "num_replies": "single" }
-----
- Instruction: 
keywords about each study
- JSON format: 
{ "name": "keywords", "type": "string", "num_replies": "multiple" }
-----
"""

################ Prompt 2: generic transform prompt when we don't have any examples
PROMPT_GET_RESULT_NO_EXAMPLE = """# Overall Description
You are a data processing assistant. I am going to give you a piece of text and a description of what to extract, and I need you to extract the data for me.  If you are not confident the data is actually present, then return empty JSON.

# Format
Extract the specified data from the given piece of text and return it in the requested JSON format. If multiple things are requested, then separate them into separate keys. The response must be valid JSON and only contain double quoted strings.

# Examples
-----
- Instruction: 
Extract all email addresses from the given text string.
- Response JSON format: 
{ "email_addresses": { "type": "string", "num_replies": "multiple" } }
- Text:
Please email me at username@gmail.com or myotheremail@uni.edu.
- Response:
{ "email_addresses": ["username@gmail.com", "myotheremail@uni.edu"] }
-----
- Instruction: 
Extract the salary from this job posting
- Response JSON format: 
{ "salary": { "type": "number", "num_replies": "single" } }
- Text:
Hello there! We are looking for a new employee to join our team. The salary is $50,000 per year.
- Response:
{ "salary": 50000 }
-----
- Instruction: 
keywords about each study
- Response JSON format: 
{
  "keywords": { "type": "string", "num_replies": "multiple" },
}
- Text:
In this paper we discuss a study on caffeine and its effect on sleep. We ran a comparative study with 48 participants across 3 age groups. Each participants consumed caffeine or a placebo
and then we measured how long it took them to fall asleep they slept. We found that on average, participants in the caffeine group took 30 minutes longer to fall asleep than the placebo group.
- Response:
{ "keywords": ["caffeine", "sleep", "comparative study"] }
-----
"""


def format_user_prompt_generic(
    user_task_description: str, response_format: dict, document: str
):
    return f"""- Instruction: 
{user_task_description}
- Response format: 
{json.dumps(response_format)}
- Text: 
{document}
- Response:
"""


################ Prompt 3: Use examples to construct prompt if we have them
PROMPT_W_EXAMPLES_BASE = """# Overall Description
You are a data processing assistant. I am going to give you a piece of text and a description of what to extract, and I need you to extract the data for me as JSON.  If you are not confident the data can be extracted then return 'n/a'.

# Format
Extract the specified data according to the format and task below and return a valid JSON object with only double quoted strings.
"""


def get_prompt_w_examples(
    user_task_description: str,
    response_format: dict,
    example_data: list,
    example_responses: list,
):
    """
    Args:
        response_format (dict): { "name": "<name>", "type": "<number|string|bool>", "num_replies": "<single|multiple>" }
        example_data (list): list of strings
        example_responses (list): list of dicts like [{"<name>": <value>}]
    """
    prompt = PROMPT_W_EXAMPLES_BASE
    prompt += f"""The response should look like:
{json.dumps(response_format)}

# Extraction Task
{user_task_description}

# Examples
-----
"""
    for d, r in zip(example_data, example_responses):
        prompt += f"- Text: \n{d}\n- Response:\n{json.dumps(r)}\n-----\n"

    print("\n\n[get_prompt_w_examples] the prompt is:::::\n\n", prompt)
    return prompt


def format_user_prompt(document):
    return f"- Text: \n{document}\n- Response:\n"


class LLMClient:
    def __init__(self, api_key: str = None):
        self.client = AsyncOpenAI(api_key=api_key)  # reads API key from env if None

    ##### ASYNC QUERY FUNCTIONS
    async def _get_format_async(self, instruction):
        """
        Get the response format in JSON. Try twice with timeout and return None if it fails.
        """
        for tryIdx in range(2):
            try:
                response = await asyncio.wait_for(
                    self.client.chat.completions.create(
                        model=MODEL_NAME,
                        response_format={"type": "json_object"},
                        messages=[
                            {"role": "system", "content": PROMPT_GET_JSON_FORMAT},
                            {"role": "user", "content": instruction},
                        ],
                        temperature=0,
                    ),
                    timeout=10,
                )
                content = response.choices[0].message.content
                r = json.loads(content)
                # TODO use pydantic to validate here
                return r
            except asyncio.TimeoutError:
                print(
                    f"[Get Format]\tTry {tryIdx}: Operation timed out after 10 seconds"
                )
                continue

        return None

    async def _do_transform_retry(self, id, system_prompt, user_prompt):
        """
        Get the transformation response from OpenAI API. Retry twice if rate limited.
        """
        print(f"[RQ {id}]\tGetting response...")

        sleepAmount = [10, 60, 60]

        for tryIdx in range(len(sleepAmount)):
            try:
                r = await asyncio.wait_for(
                    self.client.chat.completions.create(
                        model=MODEL_NAME,
                        # for newer chat.completions API
                        response_format={"type": "json_object"},
                        messages=[
                            {"role": "system", "content": system_prompt},
                            {
                                "role": "user",
                                "content": user_prompt,
                            },
                        ],
                        temperature=0,
                        # for legacy completions API
                        # prompt=f"{system_prompt}\n{user_prompt}",
                    ),
                    timeout=10,
                )
                return r
            except asyncio.TimeoutError:
                print(f"[RQ {id}]\tTry {tryIdx}: Operation timed out after 10 seconds")
                continue  # Skip to the next iteration
            except openai.RateLimitError as e:
                print(
                    f"[RQ {id}]\tTry {tryIdx}: Rate limit error, waiting for {sleepAmount[tryIdx]} sec. Error: ",
                    e,
                )
                await asyncio.sleep(sleepAmount[tryIdx])

        print(f"[RQ {id}]\tFailed to get response after {len(sleepAmount)} tries")
        return None

    async def _run_transform_generic_parallel(
        self, system_prompt, documents, user_task_description, response_format
    ):
        responses = []

        batches = generate_batches(
            system_prompt,
            [
                format_user_prompt_generic(user_task_description, response_format, d)
                for d in documents
            ],
        )

        for (
            batch_num,
            batch,
        ) in enumerate(batches):
            print(f"Running batch {batch_num + 1}/{len(batches)}")
            batch_response = await tqdm_asyncio.gather(
                *[
                    self._do_transform_retry(
                        idx,
                        system_prompt,
                        user_prompt,
                    )
                    for idx, user_prompt in enumerate(batch)
                ]
            )

            responses.extend(batch_response)
            print(f"Finished batch {batch_num + 1}/{len(batches)}")

        return responses

    async def _run_transform_w_examples_parallel(self, system_prompt, documents):
        responses = []

        batches = generate_batches(
            system_prompt,
            [format_user_prompt(d) for d in documents],
        )

        for (
            batch_num,
            batch,
        ) in enumerate(batches):
            print(f"Running batch {batch_num + 1}/{len(batches)}")
            batch_response = await tqdm_asyncio.gather(
                *[
                    self._do_transform_retry(
                        idx,
                        system_prompt,
                        user_prompt,
                    )
                    for idx, user_prompt in enumerate(batch)
                ]
            )

            responses.extend(batch_response)
            print(f"Finished batch {batch_num + 1}/{len(batches)}")

        return responses

    def parse_results(self, responses, response_format: TaskFormat):
        """
        Parses the results as JSON from a list of responses. Expects a list of openai.ChatCompletion objects.
        """
        processed_results = []
        default_response = get_default_response(response_format)
        for response in responses:
            if response:
                item = response.choices[0].message.content  # chat completions
                # item = response.choices[0].text # legacy completions
                try:
                    r = json.loads(item)
                    r = coerce_result(r, response_format)
                except Exception as e:
                    try:
                        print("Bad initial json, trying to fix")
                        item = item.replace("'", '"')
                        r = json.loads(item)
                        r = coerce_result(r, response_format)

                    except Exception as e:
                        print("Error parsing JSON: ", e)
                        r = default_response
            else:
                r = default_response

            processed_results.append(r)

        return processed_results

    ##### PUBLIC SYNC QUERY FUNCTIONS
    @timeit
    def get_response_format(self, user_task_description: str) -> dict:
        """
        Get the response format for a given user task description.

        Args:
            user_task_description (str): The description of the user's task.

        Returns:
            dict: The response format in form { "name": "<name>", "type": "<number|string|bool>", "num_replies": "<single|multiple>" }
        """
        response_format = asyncio.run(self._get_format_async(user_task_description))
        return response_format

    @timeit
    def get_transformations(
        self,
        user_task_description: str,
        response_format: TaskFormat,
        data: list,
        example_data=None,
        example_responses=None,
    ):
        """
        Runs transformation in parallel based on the given parameters.

        Args:
            user_task_description (str): The description of the user's task.
            response_format: TaskFormat
            data (list): The input data for the transformations.
            example_data (list of strings, optional): The example data for the transformations
            example_response (list, optional): The example response for the transformations. Expects data in dict with similar format to Task Format of {"<col_name": <value> }

        Returns:
            list: The parsed results of the transformations.
        """

        formatted_response_format = {
            response_format.name: {
                "type": response_format.type,
                "num_replies": response_format.num_replies,
            }
        }

        if example_data and example_responses:
            print("Running example based transformations...")
            system_prompt = get_prompt_w_examples(
                user_task_description,
                formatted_response_format,
                example_data,
                example_responses,
            )
            raw_results = asyncio.run(
                self._run_transform_w_examples_parallel(system_prompt, data)
            )

        else:
            print("Running generic transformations...")
            system_prompt = PROMPT_GET_RESULT_NO_EXAMPLE
            raw_results = asyncio.run(
                self._run_transform_generic_parallel(
                    system_prompt,
                    data,
                    user_task_description,
                    formatted_response_format,
                )
            )

        parsed_results = self.parse_results(raw_results, response_format)

        return parsed_results


def coerce_result(json_result, response_format: TaskFormat):
    """
    Coerce the result to the correct type based on the response format.
    """
    try:
        if response_format.name in json_result:
            result = json_result[response_format.name]
        elif isinstance(json_result, dict) and len(json_result) == 1:
            result = json_result[list(json_result.keys())[0]]
        else:
            result = None
    except Exception as e:
        print("Exception: ", e)
        result = None

    if response_format.num_replies == "single":
        parsed_instance = parse_instance_by_type(result, response_format.type)
    elif isinstance(result, list):
        parsed_instance = [
            parse_instance_by_type(r, response_format.type) for r in result
        ]
    else:
        parsed_instance = [parse_instance_by_type(result, response_format.type)]

    return {response_format.name: parsed_instance}


def parse_instance_by_type(result, type):
    """Force results into appropriate type."""
    if result is None and type == "bool":
        return False
    elif result is None:
        return None

    try:
        # check if default value
        if str(result).strip().lower() in ["n/a", "", "none"]:
            if type == "bool":
                return False
            return None

        if type == "number":
            return float(result)
        elif type == "bool":
            # note this will return true if not actually a bool but a real value
            return bool(result)
        elif type == "string":
            return str(result)

        return None

    except ValueError as e:
        return None


def get_default_response(response_format: TaskFormat):
    if response_format.num_replies == "single":
        return {response_format.name: None}

    return {response_format.name: []}


def generate_batches(system_prompt: str, user_prompts: List[str]):
    """
    Get batches of user prompts. Each batch should have a maximum of MAX_BATCH_SIZE user prompts AND be less than MAX_TPM tokens.
    """
    batches = []
    current_batch = []
    current_batch_tokens = 0
    total_tokens = 0

    system_prompt_tokens = get_num_tokens(system_prompt)

    for user_prompt in user_prompts:
        i_tokens = system_prompt_tokens + get_num_tokens(user_prompt)
        total_tokens += i_tokens
        if (
            len(current_batch) >= MAX_BATCH_SIZE
            or current_batch_tokens + i_tokens >= MAX_TPM
        ):
            batches.append(current_batch)
            current_batch = [user_prompt]
            current_batch_tokens = i_tokens
        else:
            current_batch.append(user_prompt)
            current_batch_tokens += i_tokens

    if current_batch:
        batches.append(current_batch)

    print(
        f"Batches generated. \nTotal documents: {len(user_prompts)}. \nTotal tokens: {total_tokens}. \nTotal batches: {len(batches)}."
    )

    return batches


def get_num_tokens(string: str):
    return len(model_encoding.encode(string))


def mainTesting():
    print("\n\n\n~~~~~~~~~~~~~~~~~~~ Starting Run.... ~~~~~~~~~~~~~~~~~~~")
    import pandas as pd
    import datetime

    llm_client = LLMClient()

    df = pd.read_parquet(
        "/Users/wepperso/workspaces/Research/TextProfileAll/TextProfiler/texturebackend/.texture_cache/raw_data/vis_papers/vis_papers.parquet"
    )

    # prompt 1: extract sports
    # df = df[(df.umap_x > 10.5) & (df.umap_y < 10)] # sports
    # user_prompt = "Which sport does this abstract have to do with? If multiple mentioned pick the main sport and if it does not have to do with sports then reply with 'n/a'."
    # response_format = TaskFormat(name="sport_name", type="string", num_replies="single")
    # call2ExResponses = [
    #     {"sport_name": "basketball"},
    #     {"sport_name": "racket"},
    #     {"sport_name": "basketball"},
    # ]

    # prompt 2: number participants
    # df = df[df.Abstract.str.contains("participants", case=False)].iloc[:30]
    # user_prompt = "Extract the number of participants this paper includes in user studies or interviews"
    # response_format = TaskFormat(
    #     name="number_participants", type="number", num_replies="single"
    # )
    # call2ExResponses = [
    #     {"number_participants": 12},
    #     {"number_participants": 0},
    #     {"number_participants": 23},
    # ]

    # prompt 3: evaluation methods
    df = df.iloc[:300]
    user_prompt = "How was this paper evaluated? Pick the single most predominant evaluation method. If no method explicitly mentioned reply with 'n/a'"
    response_format = TaskFormat(
        name="evaluation_methods", type="string", num_replies="single"
    )
    call2ExResponses = [
        {"evaluation_methods": "n/a"},
        {"evaluation_methods": "comparative study"},
        {"evaluation_methods": "User study"},
    ]

    # prompt 4
    # df = df.iloc[:30]
    # user_prompt = "Did this paper run a user study or talk to participants?"
    # response_format = TaskFormat(
    #     name="evaluation_methods", type="bool", num_replies="single"
    # )
    # call2ExResponses = [
    #     {"evaluation_methods": False},
    #     {"evaluation_methods": True},
    #     {"evaluation_methods": True},
    # ]

    data = df.Abstract.tolist()

    # STEP 1: get response format
    print("Generated response format:", response_format)

    #### VERSION 1: GENERIC RESPONSE
    # call1Data = data[3:]
    # response = llm_client.get_transformations(user_prompt, response_format, call1Data)
    # df_response = pd.DataFrame(response, columns=[response_format["name"]])
    # df_response["data"] = call1Data

    # print("Generic response:", df_response)

    #### VERSION 2: EXAMPLE RESPONSE
    call2Data = data
    call2DataExamples = data[:3]

    response2 = llm_client.get_transformations(
        user_prompt,
        response_format,
        call2Data,
        call2DataExamples,
        call2ExResponses,
    )

    df_response2 = pd.DataFrame(response2, columns=[response_format.name])

    current_time = datetime.datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
    file_path = f"RESULT_{current_time}.csv"

    df_response2.to_csv(file_path)

    print("Example response:", df_response2)

    # breakpoint()
    print("<|END|>")


if __name__ == "__main__":
    mainTesting()



from texture.models import DatasetInfo

dataset_vis_papers = DatasetInfo(
    name="vis_papers",
    primary_key={"name": "id", "type": "number"},
    origin="example",
    has_embeddings=True,
    has_projection=True,
    columns=[
        {
            "name": "Title",
            "type": "text",
        },
        {
            "name": "Abstract",
            "type": "text",
        },
        {
            "name": "word",
            "type": "categorical",
            "table_name": "vis_papers_words",
            "derived_from": "Abstract",
        },
        # {
        #     "name": "Abstract_num_words",
        #     "type": "number",
        #     "derived_from": "Abstract",
        #     "derived_how": "code",
        # },
        {
            "name": "Year",
            "type": "number",
        },
        {
            "name": "Conference",
            "type": "categorical",
        },
        {
            "name": "author",
            "type": "categorical",
            "table_name": "vis_papers_authors",
            "derived_from": "AuthorNames-Deduped",
        },
        {
            "name": "keyword",
            "type": "categorical",
            "table_name": "vis_papers_keywords",
            "derived_from": "AuthorKeywords",
        },
        {
            "name": "PaperType",
            "type": "categorical",
        },
        {
            "name": "AminerCitationCount",
            "type": "number",
        },
        {
            "name": "CitationCount_CrossRef",
            "type": "number",
        },
        {
            "name": "PubsCited_CrossRef",
            "type": "number",
        },
        {
            "name": "Award",
            "type": "categorical",
        },
        # {
        #     "name": "id",
        #     "type": "number",
        # },
        # {
        #     "name": "MODEL_sport",
        #     "type": "categorical",
        #     "derived_from": "Abstract",
        #     "derived_how": "model",
        # },
        # {
        #     "name": "AuthorNames-Deduped",
        #     "type": "categorical",
        # },
        # {
        #     "name": "AuthorKeywords",
        #     "type": "categorical",
        # },
        # {
        #     "name": "Abstract_text_length",
        #     "type": "number",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "MODEL_num_participants",
        #     "type": "number",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "MODEL_keywords",
        #     "type": "categorical",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "MODEL_summary",
        #     "type": "text",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "MODEL_has_user_study",
        #     "type": "categorical",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "DOI",
        #     "type": "categorical",
        # },
        # {
        #     "name": "Link",
        #     "type": "categorical",
        # },
        # {
        #     "name": "FirstPage",
        #     "type": "number",
        # },
        # {
        #     "name": "LastPage",
        #     "type": "number",
        # },
        # {
        #     "name": "AuthorNames",
        #     "type": "categorical",
        # },
        # {
        #     "name": "AuthorAffiliation",
        #     "type": "categorical",
        # },
        # {
        #     "name": "InternalReferences",
        #     "type": "categorical",
        # },
        #
        # {
        #     "name": "Abstract_max_word_length",
        #     "type": "number",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "Abstract_avg_word_length",
        #     "type": "number",
        #     "derived_from": "Abstract",
        # },
        # {
        #     "name": "Abstract_perc_special_chars",
        #     "type": "number",
        #     "derived_from": "Abstract",
        # },
    ],
)

dataset_airline_reviews = DatasetInfo(
    name="airline_reviews",
    primary_key={"name": "id", "type": "number"},
    origin="example",
    has_embeddings=True,
    has_projection=True,
    columns=[
        # {"name": "id", "type": "categorical"},
        # {
        #     "name": "tweet_num_characters",
        #     "type": "number",
        #     "derived_from": "tweet",
        #     "derived_how": "code",
        # },
        # {
        #     "name": "twitter_handle",
        #     "type": "categorical",
        #     "derived_from": "tweet",
        #     "derived_how": "code",
        # },
        # {
        #     "name": "is_on_hold",
        #     "type": "categorical",
        #     "derived_from": "tweet",
        #     "derived_how": "model",
        # },
        {
            "name": "word",
            "type": "categorical",
            "table_name": "tweet_words",
            "derived_from": "tweet",
        },
        {"name": "airline", "type": "categorical"},
        {"name": "airline_sentiment", "type": "categorical"},
        {"name": "airline_sentiment_confidence", "type": "number"},
        {"name": "negativereason", "type": "categorical"},
        {"name": "negativereason_confidence", "type": "number"},
        # {"name": "airline_sentiment_gold", "type": "categorical"},  # mostly null
        {"name": "name", "type": "categorical"},
        # {"name": "negativereason_gold", "type": "categorical"}, # mostly null
        {"name": "retweet_count", "type": "number"},
        {"name": "tweet", "type": "text"},
        # {"name": "tweet_coord", "type": "categorical"}, # x, y coords I think
        {"name": "tweet_created", "type": "date"},
        {"name": "tweet_location", "type": "categorical"},
        {"name": "user_timezone", "type": "categorical"},
    ],
)


EXAMPLE_DATASET_INFO = {
    "vis_papers": dataset_vis_papers,
    "airline_reviews": dataset_airline_reviews,
}

EXAMPLE_DATA_PATHS = {
    "vis_papers": {
        "datasets": {
            "vis_papers": "raw_data/vis_papers/vis_papers.parquet",
            "vis_papers_words": "raw_data/vis_papers/vis_papers_words_span.parquet",
            "vis_papers_authors": "raw_data/vis_papers/vis_papers_authors_span.parquet",
            "vis_papers_keywords": "raw_data/vis_papers/vis_papers_keywords_span.parquet",
        },
        "embeddings": {"vis_papers": "raw_data/vis_papers/vis_papers_embeddings.pt"},
    },
    "airline_reviews": {
        "datasets": {
            "airline_reviews": "raw_data/airline_reviews/airlines.parquet",
            "tweet_words": "raw_data/airline_reviews/tweet_words.parquet",
        },
        "embeddings": {"airline_reviews": "raw_data/airline_reviews/embeddings.pt"},
    },
}



import uvicorn
from typing import Dict, Union, Any, List
from multiprocessing import Process
import pandas as pd

from texture.models import (
    TextureInitArgs,
    DatasetInitArgs,
    ColumnInputInfo,
)
from texture.server import get_server
from texture.utils import is_notebook
from texture import preprocess

TEXTURE_SERVER_PROCESS = None


def run(
    data: pd.DataFrame,
    name: str = None,
    embeddings: Any = None,
    primary_key: str = None,
    column_info: List[ColumnInputInfo] = None,
    host: str = "localhost",
    port: int = 8080,
    load_example_data: bool = False,
    api_key: str = None,
):
    args = TextureInitArgs(
        data=data,
        name=name,
        embeddings=embeddings,
        primary_key=primary_key,
        column_info=column_info,
        host=host,
        port=port,
        load_example_data=load_example_data,
        api_key=api_key,
    )

    if is_notebook():
        print("Running from a notebook, starting a new process")

        global TEXTURE_SERVER_PROCESS
        if TEXTURE_SERVER_PROCESS is not None:
            print("Terminating existing server process")
            TEXTURE_SERVER_PROCESS.terminate()

        TEXTURE_SERVER_PROCESS = Process(
            target=run_server,
            args=(args,),
        )
        TEXTURE_SERVER_PROCESS.start()
        # below will block notebook from progressing past server cell
        # TEXTURE_SERVER_PROCESS.join()
    else:
        run_server(args)


def run_server(args: Union[TextureInitArgs, Dict]):
    args = TextureInitArgs(**args) if isinstance(args, dict) else args

    dsInfo, load_tables, load_embeddings = preprocess.validate_and_run_preprocess(args)

    app = get_server(
        DatasetInitArgs(
            datasetInfo=dsInfo, load_tables=load_tables, load_embeddings=load_embeddings
        ),
        load_example_data=args.load_example_data,
        api_key=args.api_key,
    )

    print(f"\n\033[1mTexture\033[0m running on http://{args.host}:{args.port}\n")
    uvicorn.run(
        app,
        host=args.host,
        port=args.port,
        log_level="warning",
        # reload=True,
    )



from fastapi import FastAPI
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import ORJSONResponse
from typing import Dict
import os
import pandas as pd
import datetime

from texture.database import init_db, populate_example_datasets, populate_dataset
from texture.models import (
    DatasetInfo,
    DuckQueryData,
    DuckQueryResult,
    VectorSearchResponse,
    TransformResponse,
    LLMTransformRequest,
    LLMTransformCommit,
    Column,
    CodeTransformRequest,
    CodeTransformCommit,
    DatasetInitArgs,
)
from texture.userTransformLLM.client import LLMClient
from texture.utils import get_type_from_response, flatten
from texture.userTransformCode.transform import (
    execute_code_and_apply_function,
)


def custom_generate_unique_id(route: APIRoute):
    """
    Replaces default generated name for TS objects with camel-cased method name.
    So python function `read_item` -> readItem
    NOTE: must ensure that route names are unique or will cause issues!
    """
    return route.name


def get_server(
    ds_init_info: DatasetInitArgs = None,
    load_example_data: bool = True,
    api_key: str = None,
) -> FastAPI:

    ### Database set up
    duckdb_conn, vectordb_conn, datasetMetadataCache = init_db()
    llm_client = LLMClient(api_key=api_key)

    if load_example_data:
        populate_example_datasets(duckdb_conn, vectordb_conn, datasetMetadataCache)

    if ds_init_info:
        populate_dataset(duckdb_conn, vectordb_conn, datasetMetadataCache, ds_init_info)

    ### Web server set up
    app = FastAPI(
        title="Texture server",
    )

    # TODO: use env variables in future for this?
    origins = [
        "http://localhost:5173",  # default vite dev
        "http://localhost:4173",  # default vite preview
        "https://dig.cmu.edu",  # deployed url
    ]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    api_app = FastAPI(
        title="Texture Backend API",
        generate_unique_id_function=custom_generate_unique_id,
        # TODO: unsure if this is necessary...
        default_response_class=ORJSONResponse,
    )

    app.mount("/api", api_app)

    app.mount(
        "/",
        StaticFiles(
            directory=os.path.dirname(os.path.realpath(__file__)) + "/frontend",
            html=True,
        ),
        name="base",
    )

    @api_app.get(
        "/status",
        response_model=str,
    )
    def root_status():
        return "hello from backend server"

    @api_app.get(
        "/all_dataset_info",
        response_model=Dict[str, DatasetInfo],
    )
    def read_dataset_info():
        """
        Get the datasets available along with a summary of their columns
        """

        return datasetMetadataCache

    @api_app.post("/duckdb_query_json", response_model=DuckQueryResult)
    def duckdb_query_json(data: DuckQueryData):
        """
        Execute a query on the database
        """
        return duckdb_conn._handle_json_message(data)

    @api_app.post("/duckdb_query_arrow")
    async def duckdb_query_arrow(data: DuckQueryData):
        """
        Execute a query on the database
        """
        return duckdb_conn._handle_arrow_message(data)

    @api_app.get("/query_embed_from_id", response_model=VectorSearchResponse)
    async def query_embed_from_id(datasetName: str, id: int):
        vector = vectordb_conn.get_embedding_from_id(datasetName, id)
        result_df = vectordb_conn.search(datasetName, vector)
        return VectorSearchResponse(
            success=True, result=result_df.to_dict(orient="records")
        )

    @api_app.get("/query_embed_from_string", response_model=VectorSearchResponse)
    async def query_embed_from_string(datasetName: str, queryString: str):
        vector = vectordb_conn.get_embedding_from_string(datasetName, queryString)
        result_df = vectordb_conn.search(datasetName, vector)
        return VectorSearchResponse(
            success=True, result=result_df.to_dict(orient="records")
        )

    @api_app.post("/fetch_llm_response_format", response_model=TransformResponse)
    def get_llm_response_format(userPrompt: str):
        try:
            task_format = llm_client.get_response_format(userPrompt)
            return TransformResponse(success=True, result=task_format)
        except Exception as e:
            print("Error in /fetch_llm_response_format::: ", str(e))
            return TransformResponse(success=False, result={"error": str(e)})

    @api_app.post("/fetch_llm_transform_result", response_model=TransformResponse)
    def get_llm_transform_result(request: LLMTransformRequest):
        try:
            results = llm_client.get_transformations(
                request.userPrompt,
                request.taskFormat,
                request.columnData,
                request.exampleData,
                request.exampleResponse,
            )
            parsed_results = [r[request.taskFormat.name] for r in results]
            return TransformResponse(success=True, result=parsed_results)
        except Exception as e:
            print("ERROR in /fetch_llm_transform_result ", str(e))
            return TransformResponse(success=False, result={"error": str(e)})

    @api_app.post("/commit_llm_transform_result", response_model=TransformResponse)
    def commit_llm_transform_result(request: LLMTransformCommit):
        try:
            new_col_name = request.taskFormat.name

            # Step 1: get data
            all_data_df = duckdb_conn.connection.execute(
                f'SELECT "id", "{request.columnName}" from "{request.tableName}"'
            ).df()
            transform_data = pd.merge(
                all_data_df, pd.DataFrame({"id": request.applyToIndices}), on="id"
            )

            # Step 2: get results and turn into flat array
            results = llm_client.get_transformations(
                request.userPrompt,
                request.taskFormat,
                transform_data[request.columnName],
                request.exampleData,
                request.exampleResponse,
            )

            # Step 3: format with correct indices transform
            processed_results = [r[new_col_name] for r in results]
            processed_df = pd.DataFrame(
                {new_col_name: list(processed_results), "id": transform_data["id"]}
            )

            colType = get_type_from_response(request.taskFormat.type)

            if request.taskFormat.num_replies == "multiple":
                newTableDf = flatten(processed_df[new_col_name], idColName="id")
                if colType == "number":
                    newTableDf[new_col_name] = pd.to_numeric(newTableDf[new_col_name])
                newTableName = new_col_name + "_table"
                duckdb_conn.load_dataframe(newTableName, newTableDf)
                newColSchema = Column(
                    name=new_col_name,
                    type=colType,
                    table_name=newTableName,
                    derived_from=request.columnName,
                    derived_how="model",
                )
            else:
                if colType == "number":
                    processed_df[new_col_name] = pd.to_numeric(
                        processed_df[new_col_name]
                    )

                all_merged = pd.merge(all_data_df, processed_df, on="id", how="left")

                duckdb_conn.add_column(
                    request.tableName, new_col_name, all_merged[new_col_name]
                )
                newColSchema = Column(
                    name=new_col_name,
                    type=colType,
                    derived_from=request.columnName,
                    derived_how="model",
                )

            datasetMetadataCache[request.tableName].columns.insert(0, newColSchema)

            return TransformResponse(success=True, result=[])

        except Exception as e:
            print("Error in /commit_llm_transform_result::: ", str(e))
            return TransformResponse(success=False, result={"error": str(e)})

    @api_app.post("/fetch_code_transform_result", response_model=TransformResponse)
    def get_code_transform_result(request: CodeTransformRequest):
        try:
            df = pd.DataFrame({"sample": request.columnData})
            results = execute_code_and_apply_function(request.codeString, df["sample"])
            if results is None:
                raise Exception("No results returned from code execution")

            return TransformResponse(success=True, result=list(results))

        except Exception as e:
            print("Exception running user code: ", e)
            return TransformResponse(success=False, result={"error": str(e)})

    @api_app.post("/commit_code_transform_result", response_model=TransformResponse)
    def commit_code_transform_result(request: CodeTransformCommit):
        new_col_name = request.taskFormat.name

        # Step 1: get data
        all_data_df = duckdb_conn.connection.execute(
            f'SELECT "id", "{request.columnName}" from "{request.tableName}"'
        ).df()
        transform_data = pd.merge(
            all_data_df, pd.DataFrame({"id": request.applyToIndices}), on="id"
        )

        # Step 2: do transform
        try:
            results = execute_code_and_apply_function(
                request.codeString, transform_data[request.columnName]
            )
            if results is None:
                raise Exception("No results returned from code execution")
        except Exception as e:
            print("Exception running user code: ", e)
            return TransformResponse(success=False, result={"error": str(e)})

        # Step 3: format with correct indices transform
        processed_df = pd.DataFrame(
            {new_col_name: list(results), "id": transform_data["id"]}
        )
        colType = get_type_from_response(request.taskFormat.type)

        if request.taskFormat.num_replies == "multiple":
            newTableDf = flatten(processed_df[new_col_name], idColName="id")
            if colType == "number":
                newTableDf[new_col_name] = pd.to_numeric(newTableDf[new_col_name])
            newTableName = new_col_name + "_table"
            duckdb_conn.load_dataframe(newTableName, newTableDf)
            newColSchema = Column(
                name=new_col_name,
                type=colType,
                table_name=newTableName,
                derived_from=request.columnName,
                derived_how="code",
            )
        else:
            if colType == "number":
                processed_df[new_col_name] = pd.to_numeric(processed_df[new_col_name])

            all_merged = pd.merge(all_data_df, processed_df, on="id", how="left")

            duckdb_conn.add_column(
                request.tableName, new_col_name, all_merged[new_col_name]
            )
            newColSchema = Column(
                name=new_col_name,
                type=colType,
                derived_from=request.columnName,
                derived_how="code",
            )

        datasetMetadataCache[request.tableName].columns.insert(0, newColSchema)
        return TransformResponse(success=True, result=[])

    @api_app.post("/save_to_file", response_model=bool)
    def save_database_to_file(table_name: str):

        all_table_names = set([table_name])
        # get all table names
        for col in datasetMetadataCache[table_name].columns:
            if col.table_name is not None:
                all_table_names.add(col.table_name)

        print("Saving tables: ", all_table_names)

        for t_name in all_table_names:
            current_time = datetime.datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
            file_path = f"{t_name}_{current_time}.parquet"
            duckdb_conn.write_table_to_file(t_name, file_path)

        return True

    return app



from typing import List, Dict, Union, Literal, Any, Optional
from pydantic import BaseModel
import pandas as pd


#### Internal API models between frontend and backend

DataType = Literal["text", "number", "date", "categorical"]


class Column(BaseModel):
    name: str
    type: DataType
    derived_from: Optional[str] = None
    # for join datasets, if None assumed to be in main table
    table_name: Optional[str] = None
    derived_how: Optional[Literal["model", "code"]] = None


class DatasetInfo(BaseModel):
    name: str
    columns: List[Column]
    primary_key: Column
    origin: Literal["example", "uploaded"]
    has_embeddings: bool = False
    has_projection: bool = False


class ColumnSummary(BaseModel):
    column_name: str
    column_type: str
    min: str
    max: str
    approx_unique: str
    avg: str = None
    std: str = None
    q25: str = None
    q50: str = None
    q75: str = None
    count: int
    null_percentage: str


class DuckQueryData(BaseModel):
    uuid: str
    sql: str
    type: Literal["arrow", "exec", "json"]
    buffers: list = []


class ErrorResponse(BaseModel):
    uuid: str
    type: Literal["error"] = "error"
    error: str


class ExecResponse(BaseModel):
    uuid: str
    type: Literal["exec"] = "exec"


class JsonResponse(BaseModel):
    uuid: str
    type: Literal["json"] = "json"
    result: Union[List, Dict]  # any valid JSON


DuckQueryResult = Union[ExecResponse, JsonResponse, ErrorResponse]


class DatasetUploadResponse(BaseModel):
    success: bool
    message: str
    datasetSchema: DatasetInfo = None


class DatasetVerifyResponse(BaseModel):
    success: bool
    message: str


class DatasetTokenizeResponse(BaseModel):
    success: bool
    message: str


class VectorSearchResponse(BaseModel):
    success: bool
    result: List[Dict[str, Any]]


class TransformResponse(BaseModel):
    success: bool
    result: Union[List, Dict]


# TODO rename this to TranformSchema
class TaskFormat(BaseModel):
    name: str
    type: Literal["number", "string", "bool"]
    num_replies: Literal["single", "multiple"]


class LLMTransformRequest(BaseModel):
    userPrompt: str
    taskFormat: TaskFormat
    columnData: List[str]
    exampleData: Optional[List[str]] = None
    exampleResponse: Optional[List[Dict[str, Any]]] = (
        None  # list of strings, numbers, or bools
    )


class LLMTransformCommit(BaseModel):
    userPrompt: str
    taskFormat: TaskFormat
    columnName: str
    tableName: str
    exampleData: List[str]
    exampleResponse: List[Dict[str, Any]]
    applyToIndices: List[int]


class CodeTransformRequest(BaseModel):
    codeString: str
    taskFormat: TaskFormat
    columnData: List[str]


class CodeTransformCommit(BaseModel):
    codeString: str
    taskFormat: TaskFormat
    columnName: str
    tableName: str
    applyToIndices: List[int]


#### public API for running


DataFrameType = Any  # Future: make sure this is pd.DataFrame or huggingface dataset


class DatasetInitArgs(BaseModel):
    datasetInfo: DatasetInfo
    load_tables: Dict[str, DataFrameType]
    load_embeddings: Optional[Dict[str, Any]] = None


class ColumnInputInfo(Column):
    table_data: Optional[DataFrameType] = None


class TextureInitArgs(BaseModel):
    data: DataFrameType  # TODO: also support huggingface dataset
    name: Optional[str] = None
    embeddings: Optional[Any] = None  # torch.tensor or np.ndarray
    primary_key: Optional[str] = None
    column_info: Optional[List[ColumnInputInfo]] = None
    host: Optional[str] = "localhost"
    port: Optional[int] = 8080
    api_key: Optional[str] = None
    load_example_data: Optional[bool] = False



import random
import torch
import numpy as np
from typing import List

from texture import preprocess
from texture.models import TextureInitArgs, DatasetInfo, Column, ColumnInputInfo


def validate_and_run_preprocess(args: TextureInitArgs):  # -> (DatasetInfo, Dict)
    df = args.data
    name = args.name
    column_info = args.column_info
    sanitized_embeddings = None
    has_embeddings = False
    has_projection = False

    if not name:
        name = "dataset_" + str(random.randint(1000, 9999))

    if not column_info:
        column_info = preprocess.get_data_types(df)

    # make primary key if none provided (if provided make sure is unique)
    pk_name = args.primary_key
    if not (pk_name and pk_name in df.columns and df[pk_name].is_unique):
        if "id" in df.columns and df["id"].is_unique:
            pk_name = "id"
        else:
            # print("No valid primary key found, creating a new one...")
            df = df.reset_index(drop=True)
            df = df.reset_index().rename(columns={"index": "id"})
            pk_name = "id"
            column_info.append(ColumnInputInfo(name=pk_name, type="number"))

    # check if embeddings
    if args.embeddings is not None:

        sanitized_embeddings = verify_embeddings(args.embeddings)
        has_embeddings = True

        if not ("umap_x" in df.columns and "umap_y" in df.columns):
            print(
                "Did not find projection included (no umap_x, umap_y columns), calculating now..."
            )
            projection = preprocess.get_projection(sanitized_embeddings)

            df["umap_x"] = projection[:, 0]
            df["umap_y"] = projection[:, 1]
            has_projection = True

    # or just projection of embeddings
    if "umap_x" in df.columns and "umap_y" in df.columns:
        has_projection = True

    # populate dataset Info
    dsInfo, load_tables = create_ds_info(
        name,
        column_info,
        pk_name,
        has_embeddings,
        has_projection,
    )
    load_tables[name] = df
    load_embeddings = None
    if sanitized_embeddings is not None:
        load_embeddings = {name: sanitized_embeddings}

    return dsInfo, load_tables, load_embeddings


def verify_embeddings(e) -> np.ndarray:
    if isinstance(e, np.ndarray):
        return e
    if isinstance(e, torch.Tensor):
        return e.cpu().numpy()

    raise ValueError(
        "Embeddings must be a numpy.ndarray or torch.Tensor. Received: ", type(e)
    )


def create_ds_info(
    name: str,
    column_info: List[ColumnInputInfo],
    pk_name: str,
    has_embeddings: bool,
    has_projection: bool,
):

    pk_info = Column(name=pk_name, type="number")  # backup if not found

    cleaned_cols = []
    load_tables = {}

    for c_info in column_info:
        if c_info.name == pk_name:
            pk_info = c_info

        if c_info.table_name is not None and c_info.table_data is None:
            raise ValueError(f"Column {c_info.name} has table_name but no table_data")

        table_name = None
        if c_info.table_data is not None:

            # validate the passsed dataframe
            cols = c_info.table_data.columns
            if pk_name not in cols:
                raise ValueError(
                    f"Primary key {pk_name} not found in column table:", c_info
                )
            if c_info.name not in cols:
                raise ValueError(
                    f"Column name {c_info.name} not found in column table:",
                    c_info,
                )

            # TODO: check if span_start and span_end are in columns maybe

            if c_info.table_name:
                table_name = c_info.table_name
            else:
                table_name = f"{c_info.name}_{c_info.derived_from}"

            load_tables[table_name] = c_info.table_data

        # make Column without table_data from ColumnInputInfo
        c = Column(
            name=c_info.name,
            type=c_info.type,
            derived_from=c_info.derived_from,
            table_name=table_name,
            derived_how=c_info.derived_how,
        )

        cleaned_cols.append(c)

    return (
        DatasetInfo(
            name=name,
            primary_key=pk_info,
            origin="uploaded",
            columns=cleaned_cols,
            has_embeddings=has_embeddings,
            has_projection=has_projection,
        ),
        load_tables,
    )



from texture.runner import run
import pandas as pd


def dev_run():
    """For launching server from poetry"""
    # N.B frontend dev server hard coded for localhost:8080
    df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "text": [
                "Hello world how are you?",
                "I'm fine how about you?",
                "The weather is great today.",
            ],
            "label": [0, 1, 0],
            "umap_x": [0.1, 0.2, 0.3],
            "umap_y": [0.4, 0.5, 0.6],
        }
    )
    run(df, load_example_data=True)


if __name__ == "__main__":
    dev_run()



import umap
import sentence_transformers
import numpy as np
from pathlib import Path
import torch

from texture.preprocess.utils import save_embeddings

# see https://www.sbert.net/docs/pretrained_models.html for all models
# best model overall for english: "all-mpnet-base-v2"
# best semantic similarity across languages: "distiluse-base-multilingual-cased-v1"
# best model for Bitext Mining i.e. translation: "LaBSE"


def calculate_embeddings(
    col: np.ndarray, model_name="all-mpnet-base-v2", useTensor=True
):
    """
    Calcs the embeddings.
    TODO: My understanding is the sentence_transformers library cuts off inputs longer than
    384 words. Need to fix that for longer docs probably by chunking text into longest amount and
    then averaging the resulting embeddings
    """
    print("Calculating embeddings with ", model_name)
    model = sentence_transformers.SentenceTransformer(model_name)
    e = model.encode(col, convert_to_tensor=useTensor)
    print("Created embedding of shape", e.shape, "with", model_name)
    return e


def get_projection(vector):
    # alt metric is "cosine"
    # NOTE: for some reason this breaks on intel mac laptop if n_jobs is not 1 so dont change
    print("Calculating UMAP projection")
    return umap.UMAP(metric="euclidean", n_components=2, n_jobs=1).fit_transform(
        vector,
    )


def get_embeddings_and_projection(
    col_matrix: np.ndarray, base_path: str = None, model_name="all-mpnet-base-v2"
):
    embeddings = calculate_embeddings(col_matrix, model_name)

    local_embeds = (
        embeddings.cpu().numpy() if isinstance(embeddings, torch.Tensor) else embeddings
    )

    if base_path:
        embed_path = Path(base_path) / "embeddings.pt"
        save_embeddings(local_embeds, embed_path)

    projection = get_projection(local_embeds)
    return embeddings, projection
